----------------------------------------------------------------------------------------------------
----   .                                                                 ----
----  parseCmdline   Cmdline           ----
----    .                                                                      ----
----      ,  find_archives              ----
----      .                                                                      ----
----          :                       ----
---- *        runArchiveCreate     ArcCreate   ( a/f/m/u/j/d) -
---- *           -  runArchiveExtract  -         ArcExtract  ( t/e/x)    ----
---- *    -  runArchiveList     -         ArcList     ( l/v)      ----
----          .  ----
----                                                                                            ----
----           :                      ----
----   ArhiveFileList   -                                  ----
----   ArhiveStructure  -                                            ----
----   ByteStream       -                  ----
----   Compression      -    ,    CRC           ----
----   CUI              -        :)           ----
----   Errors           -                                       ----
----   FileInfo         -                         ----
----   Files            -                           ----
----   Process          -          ----
----   Utils            -                                 ----
----------------------------------------------------------------------------------------------------
module Main where

import Prelude hiding (catch)
import Control.Concurrent
import Control.Exception
import Control.Monad
import Data.Char
import Data.List
import System.Directory
import System.Environment
import System.Exit

import Utils
import Process
import Errors
import Files
import FileInfo
import CUI
import Cmdline
import ArcCreate
import ArcList
import ArcExtract


-- |  
main         =  doMain =<< getArgs
-- |     
arc cmdline  =  doMain (words cmdline)

-- |        
doMain args  = do
  --handleCtrlBreak (return ()) $ do      --  ^Break    ThreadKilled
  do
    uiStartProgram                      --  UI
    commands <- parseCmdline args       --        
    mapM_ run commands                  --    
    uiDoneProgram                       --  UI
  `catch`
    \e -> do case (userErrors e, asyncExceptions e) of
               (Just str, _)                  -> putStr ("\n"++str)
               (Nothing,  Just ThreadKilled)  -> putStr "\nProgram terminated!"
               _                              -> throwIO e
             exitWith (ExitFailure 1)


-- |         
run command @ Command {
                cmd_name         = cmd
              , opt_scan_subdirs = scan_subdirs
              } = do
  case (cmd) of
    "create" -> find_archives  False           run_add     command
    "a"      -> find_archives  False           run_add     command
    "f"      -> find_archives  False           run_add     command
    "m"      -> find_archives  False           run_add     command
    "u"      -> find_archives  False           run_add     command
    "j"      -> find_archives  False           run_join    command
    "cw"     -> find_archives  False           run_cw      command
    "y"      -> find_archives  scan_subdirs    run_copy    command
    "s"      -> find_archives  scan_subdirs    run_copy    command
    "c"      -> find_archives  scan_subdirs    run_copy    command
    "k"      -> find_archives  scan_subdirs    run_copy    command
    "d"      -> find_archives  scan_subdirs    run_delete  command
    "e"      -> find_archives  scan_subdirs    run_extract command
    "x"      -> find_archives  scan_subdirs    run_extract command
    "t"      -> find_archives  scan_subdirs    run_test    command
    "l"      -> find_archives  scan_subdirs    run_list    command
    "v"      -> find_archives  scan_subdirs    run_list    command
    _ -> registerError$ UNKNOWN_CMD cmd aLL_COMMANDS

-- | ,    arcspec,        
find_archives scan_subdirs   --     
              run_command
              command @ Command {cmd_arcspec = arcspec} = do
  uiStartCommand command   --    
  arclist <- if scan_subdirs || is_wildcard arcspec
             then find_files scan_subdirs arcspec >>== map diskName
             else return [arcspec]
  results <- (`mapM` arclist) $ \arcname -> do
      --    -ad,          ( )
    let add_dir  =  if opt_add_dir command  then (`joinDirFilename` getFilePrefix arcname)  else id
    run_command command { cmd_arcspec = error "cmd_arcspec undefined"  -- cmd_arcspec    
                        , cmd_arclist = arclist
                        , cmd_arcname = arcname
                        , opt_disk_basedir = add_dir (opt_disk_basedir command)
                        }
  uiDoneCommand command results   --        


-- |   : create, a, f, m, u
run_add cmd @ Command { cmd_filespecs     =  filespecs
                      , opt_scan_subdirs  =  scan_subdirs
                      , opt_arc_basedir   =  arc_basedir
                      , opt_disk_basedir  =  disk_basedir
                      , opt_include_dirs  =  include_dirs
                      } = do
  --     ,        
  --             ,
  --     
  let filter_f  =  all_functions [not.excluded_f cmd, not.overwrite_f cmd]
      diskfiles =  find_and_filter_files scan_subdirs include_dirs filter_f arc_basedir disk_basedir filespecs
  runArchiveCreate cmd{ cmd_diskfiles = diskfiles }  -- ,     

-- |  : j
run_join cmd @ Command { cmd_filespecs     =  filespecs
                       , opt_scan_subdirs  =  scan_subdirs
                       , opt_disk_basedir  =  disk_basedir
                       , opt_noarcext      =  noarcext
                       } = do
  --     ,        
  --             ,
  --     
  let filter_f  =  all_functions [not.excluded_f cmd, not.overwrite_f cmd]
      arcspecs  =  map (addArcExtension noarcext) filespecs   --       (".arc")
      arcnames  =  find_and_filter_files scan_subdirs (Just False) filter_f "" disk_basedir arcspecs >>== map diskName
  runArchiveCreate cmd{ cmd_added_arcnames = arcnames }  --   

-- |     : y, c, k
run_copy cmd = do
  --     ,    ,       "-x"
  let arc_filter  =  all_functions [included_f cmd, not.excluded_f cmd]
  runArchiveCreate cmd{ cmd_arc_filter = arc_filter }  --      

-- |   : d
run_delete cmd = do
  --     ,    ,       "-x"
  let arc_filter  =  not.all_functions [included_f cmd, not.excluded_f cmd]
  runArchiveCreate cmd{ cmd_arc_filter = arc_filter }  --      

-- |   : e, x
run_extract cmd = do
  --   ,    ,      "-x".
  --             ,
  --        
  let arc_filter  =  all_functions [included_f cmd, not.excluded_f cmd, not.overwrite_f cmd]
  runArchiveExtract cmd{ cmd_arc_filter = test_dirs cmd arc_filter }  --      

-- |  : t
run_test cmd = do
  --   ,    ,       "-x"
  let arc_filter  =  all_functions [included_f cmd, not.excluded_f cmd]
  runArchiveExtract cmd{ cmd_arc_filter = test_dirs cmd arc_filter }  --      

-- |   : l, v
run_list cmd = do
  --     ,    ,       "-x"
  let arc_filter  =  all_functions [included_f cmd, not.excluded_f cmd]
  runArchiveList cmd{ cmd_arc_filter = test_dirs cmd arc_filter }  --      

-- |     : cw
run_cw = runCommentWrite

{-# NOINLINE find_archives #-}
{-# NOINLINE run_add #-}
{-# NOINLINE run_join #-}
{-# NOINLINE run_copy #-}
{-# NOINLINE run_delete #-}
{-# NOINLINE run_extract #-}
{-# NOINLINE run_test #-}
{-# NOINLINE run_list #-}


----------------------------------------------------------------------------------------------------
----    ,   ----------------------------------------
----------------------------------------------------------------------------------------------------

-- |  ,    
included_f  cmd  =  match_filespecs (opt_match_with cmd) (cmd_filespecs cmd) . fiFilteredName

-- |  ,   "-x"
excluded_f  cmd  =  (opt_match_excluded cmd) . fiFilteredName

-- |  ,     
--   ,   
-- (  -     ,  ,   :)
overwrite_f cmd  =  in_arclist_or_temparc (cmd_arclist cmd) . fiDiskName

in_arclist_or_temparc arclist filename =
  fpFullname filename `elem` arclist
  || (   temparc_prefix `isPrefixOf` fpBasename filename
      && temparc_suffix `isSuffixOf` fpBasename filename)

-- |     `filter_f`        `cmd`
test_dirs cmd filter_f fi  =  if fiIsDir fi
                              then opt_x_include_dirs cmd
                              else filter_f fi

