----------------------------------------------------------------------------------------------------
----    .                                                              ----
----    ,      :    ----
----   create/a/f/m/u/d/j.                                                                      ----
----  runArchiveCreate   ,      , ----
----         ,   ,     ----
----         .      .          ----
----------------------------------------------------------------------------------------------------
module ArcCreate where

import Prelude hiding (catch)
import Control.Concurrent
import Control.Exception
import Control.Monad
import Data.IORef
import Data.List
import Foreign.C.Types
import Foreign.Ptr
import Foreign.Marshal.Alloc
import Foreign.Marshal.Pool
import Foreign.Marshal.Utils
import System.Directory
import System.Environment
import System.Exit
import System.IO

import Utils
import Files
import Process
import Errors
import FileInfo
import Cmdline
import Compress
import CUI
import ArhiveStructure
import ArhiveFileList
import ArcExtract


-- |  / 
runArchiveCreate command @ Command {            --    :
      cmd_name           = cmd                  --    
    , cmd_arcname        = arcname              --    ,   
    , cmd_arc_filter     = arc_filter           --        
    , cmd_added_arcnames = find_added_arcnames  --     
    , cmd_diskfiles      = find_diskfiles       --   ,     
    , opt_arccmt_file    = arccmt_file          --   ,       
    } = do
  --   : find_files |> buffer 10_000 |> write_to_archive
  --  "create"     
  when (cmd=="create")$  do tryFileRemove arcname
  --      
  uiStartArchive command (encode_method$ opt_data_compressor command)
  statMark "Starting"
  --    (   -   "j")  
  diskfiles      <- find_diskfiles
  added_arcnames <- find_added_arcnames
  statMark "Scanning"
  --     ,   
  main_archive    <-  phantomArchiveReadInfo command "" "" arc_filter  arcname
  added_archives  <-  mapM  (archiveReadInfo command "" "" arc_filter) added_arcnames
  let input_archives = main_archive:added_archives     --    
      closeInputArchives = do mapM_ arcClose input_archives
  when (arcLocked main_archive) $
    registerError$ GENERAL_ERROR "can't modify archive locked with -k"
  --    
  aromment  <-  getAromment arccmt_file input_archives
  --statMark "Reading archive structure"
  --     
  let sorted_diskfiles  =  sort_files command diskfiles
  statMark$ "Sorting "++ show (length sorted_diskfiles)++" files"
  --   ,      ,
  --       ,     (
  -- "j")   ,    .      .
  files_to_archive <- join_lists main_archive added_archives sorted_diskfiles command
  statMark "Joining filelists"
  if null files_to_archive              --        
    then do registerWarning NOFILES     --     
            closeInputArchives          --      
            tryFileRemove arcname       --     ,      (, "Arc d archive *")
            return (1,0,0,0)
    else do
  --         
  results <- ref (error "results undefined")
  --  mtime    -tk
  old_arc_exist <- fileExist arcname
  arc_time <- if old_arc_exist  then getFileDateTime arcname  else return (error "arc_time undefined")
  --    -tl               .
  --      create_archive_structure_PROCESS   `find_last_time`.
  --        ,   ,       .
  --          .
  last_time <- ref aMINIMAL_POSSIBLE_DATETIME
  let find_last_time dir = do
        modifyIORef last_time $ \last_time' ->
          maximum (last_time' : map (fiTime.cfFileInfo) dir)
  let processDir = if (opt_time_to_last command)  then find_last_time  else (\dir->return())
  --      
  uiStartProcessing (map cfFileInfo files_to_archive)

  --           ,  ,    -  
  tempfile_wrapper arcname $  \temp_arcname -> do
    --      
    bracket (archiveCreate temp_arcname) (archiveClose) $  \archive -> do
      writeSFX (opt_sfx command) archive main_archive    --  SFX-    
      withPool $ \pool -> do
        --   -   ,    :
        --         
        --          
        let read_files          =  create_archive_structure_AND_read_files_PROCESS command archive files_to_archive processDir pool aromment
            compress_AND_write  =  compress_AND_write_to_archive_PROCESS archive
        backdoor <- newChan  --       
        runP( read_files backdoor |>>> compress_AND_write backdoor )
    closeInputArchives                       --   
    --statMark "Writing"
    uiDoneArchive  >>=  writeIORef results   --            
    when (opt_test command) $ do             --    "-t"
      testArchive command temp_arcname       --        

  when (opt_keep_time command && old_arc_exist) $ do     --    -tk      
    setFileDateTime arcname arc_time                     --     mtime 
  when (opt_time_to_last command) $ do                   --    -tl
    setFileDateTime arcname =<< val last_time            --     &    &      
  renameArchiveAsSFX arcname command                     --  ,          SFX-
  val results                                            --    


-- |     ,   -z,
--     ,     
getAromment arccmt_file input_archives = do
  case arccmt_file of               --      "-z":
    ""   -> uiInputArcComment       --      stdin
    "-"  -> return ""               --     
    "--" ->                         --      ( )
            do let comment = join_with "\n" (map arcComment input_archives)
               uiPrintArcComment comment
               return comment
    _    -> fileGet arccmt_file     --        

-- | SFX-    
writeSFX sfxname archive old_archive = do
  let oldArchive = arcArchive old_archive
      oldSFXSize = arcSFXSize old_archive
  case sfxname of                                      --      "-sfx":
    "-"      -> return ()                              --     sfx-
    "--"     -> unless (arcPhantom old_archive) $ do   --    sfx    ( )
                  archiveCopyData oldArchive 0 oldSFXSize archive
    filename -> bracket (archiveOpen sfxname              --     sfx   
                          `catch` (\e -> registerError$ GENERAL_ERROR$ "can't open SFX module "++sfxname))
                        (archiveClose)
                        (\sfxfile -> do size <- archiveGetSize sfxfile
                                        archiveCopyData sfxfile 0 size archive)

-- |     ,         SFX-
renameArchiveAsSFX arcname command = do
  let exe = aDEFAULT_SFX_EXTENSION
      arc = aDEFAULT_ARC_EXTENSION
  let changeExt = case (opt_noarcext command, opt_sfx command) of
        (True, _)     -> id                     --   ,    --noarcext
        (_   , "--")  -> id                     --       "-sfx"
        (_   , "-")   -> replaceAtEnd exe arc   --  "-sfx-"    ".arc"
        _             -> replaceAtEnd arc exe   --  "-sfx..."    ".exe"
  let newname = changeExt arcname
  when (newname/=arcname) $ do
    putStr$ "\nRenaming "++arcname++" to "++newname
    fileRename arcname newname

-- |    ,      `temp_arcname`
testArchive command temp_arcname = do
  let test_command = command{ cmd_name         = "t"           -- 
                            , cmd_arcname      = temp_arcname  --   
                            , opt_arc_basedir  = ""            --  
                            , opt_disk_basedir = ""            -- ...
                            , cmd_arc_filter   = const True    -- ...
                            }
  uiStartCommand  test_command
  results      <- runArchiveExtract test_command
  uiDoneCommand   test_command [results]
  uiStartCommand  command



---------------------------------------------------------------------------------------------------
----     ----------------------------------------------------------------
---------------------------------------------------------------------------------------------------

-- |,     
data Instruction
  =   CompressData BlockType Compressor Bool  --     
  |     FileStart FileInfo                    --       
  |       DataChunk (Ptr CChar) Int           --          
  |     CorrectTotals FileSize                --      Total Size  UI
  |     FakeFiles [CompressedFile]            --           UI
  |   CopySolidBlock [CompressedFile]         --    -    
  |   DataEnd                                 --     
  |   Directory [CompressedFile]              --            
  | TheEnd                                    --   

-- |    :     ,    
notDataEnd DataEnd = False
notDataEnd _       = True
notTheEnd  TheEnd  = False
notTheEnd  _       = True


-- |    // 
writeBlock pipe blockType compressor just_copy action = do
  sendP pipe (CompressData blockType compressor just_copy)
  directory <- action
  sendP pipe  DataEnd
  sendP pipe (Directory directory)


-- |,    -   
--    ,      -.
--         -     
--     .
--
create_archive_structure_AND_read_files_PROCESS command archive files_to_archive processDir pool aromment backdoor pipe = do
    --        
  decompress_pipe <- runAsyncP( decompress_PROCESS (return) )
    --       
  heap                     <-  pooledMallocBytes pool (opt_cache command)
  (getBlock, shrinkBlock)  <-  memoryAllocator   heap (opt_cache command) aBUFFER_SIZE 256 (receive_backP pipe)

  let --      ,      UI
      startFile  fi         =  sendP pipe (FileStart fi)
      correctTotals addme   =  when (addme/=0)$  do sendP pipe (CorrectTotals addme)
      --          
      receiveBuf            =  do block <- getBlock; return (block, aBUFFER_SIZE)
      sendBuf buf size len  =  do shrinkBlock buf len
                                  when (len>0)$  do sendP pipe (DataChunk buf len)
      --    read_file
      bufOps                =  (startFile, correctTotals, receiveBuf, sendBuf)

  let --               
      writeControlBlock blockType compressor action = do
        if (opt_nodir command)   --  "--nodir"        -     
          then return$ error "Tried to use value of writeControlBlock when \"--nodir\""
          else do
        writeBlock pipe blockType compressor False $ do         --     
          action; return []
        (block, [])  <-  getP backdoor                          --      
        writeBlock pipe DESCR_BLOCK aNO_COMPRESSION False $ do  --   
          archiveWriteBlockDescriptor block receiveBuf sendBuf; return []
        (_,     [])  <-  getP backdoor                          --     
        return block                                            --   

  let --     -     (splitByType)
      --    opt_group_data   - (splitByCrit).
      --  ,    (grzip, lzp) -        
      splitToSolidBlocks xs = let compressor = data_compressor xs
                                  bs         = maximum (map freearcGetBlockSize compressor)
                                  add_crit   = if bs>0  then (GroupByBlockSize bs :)  else id
                              in case compressor of
                                   [] -> [xs]  --  "-m0"      -
                                   _  -> splitByCrit (add_crit$ opt_group_data command) xs
      splitByCrit crit = concatMap (splitBy crit) . splitByType
      splitByType      = partitionList (opt_types_count command) (opt_find_type command.cfFileInfo)
      --       
      data_compressor (file:_) =  snd (opt_data_compressor command !! opt_find_type command (cfFileInfo file))
      --      ( HEADER_BLOCK,    )
      dir_compressor  =  snd$ head$ opt_dir_compressor command
      -- True,       
      lock_archive    =  opt_lock_archive command


  ------------------------------------------------------------------------------------------------
  --  ,     ------------------------------------
  ------------------------------------------------------------------------------------------------
  --         (HEADER_BLOCK)
  header_block  <-  writeControlBlock HEADER_BLOCK aNO_COMPRESSION $ do
                      archiveWriteHeaderBlock receiveBuf sendBuf

  --      ,         ,
  --      
  directory_blocks  <-  (`mapM` splitBy (opt_group_dir command) files_to_archive)
    ( \filesInOneDirectory -> do       --  ,      

      --    -     .   - mapM (print.map (fpFullname.fiDiskName.cfFileInfo)) (splitToSolidBlocks filesInOneDirectory)
      datablocks  <-  (`mapM` splitToSolidBlocks filesInOneDirectory)
        ( \filesInOneDataBlock -> do     --     -
          let --      -     ,
              --       
              compressor = map (freearcLimitDictionary$ clipToMaxInt totalBytes)
                               (data_compressor filesInOneDataBlock)
              totalBytes = sum$ map (fiSize.cfFileInfo) filesInOneDataBlock
              -- True,   -    -   
              copy_solid_block = isWholeSolidBlock filesInOneDataBlock
          --  -        
          writeBlock pipe DATA_BLOCK compressor copy_solid_block $ do
            dir <- --  -      ,     +
                   if copy_solid_block then do
                     sendP pipe (CopySolidBlock filesInOneDataBlock)
                     return filesInOneDataBlock   --     cfArcBlock/cfPos   ,    -      
                   --   FAKE_COMPRESSION,     
                   else if (compressor==[aFAKE_COMPRESSION]) then do
                     sendP pipe (FakeFiles filesInOneDataBlock)
                     return$ map (makeCompressedFile 0 . cfFileInfo) filesInOneDataBlock
                   else do
                     mapMaybeM (read_file command bufOps decompress_pipe) filesInOneDataBlock
            processDir dir   --  ,  ,     (    -tl)
            return dir
        )

      --   write_to_archive          .
      --          ...
      blocks_info  <-  replicateM (length datablocks) (getP backdoor)
      -- ...     ,       
      arcpos <- archiveGetPos archive
      --            
      writeControlBlock DIR_BLOCK dir_compressor $ do
        archiveWriteDir blocks_info arcpos receiveBuf sendBuf
    )

  arcpos <- archiveGetPos archive
  --    (FOOTER_BLOCK),        
  writeControlBlock FOOTER_BLOCK dir_compressor $ do
    archiveWriteFooterBlock (header_block:directory_blocks) lock_archive aromment arcpos receiveBuf sendBuf
  --    
  sendP decompress_pipe Nothing
  joinP decompress_pipe
  --   ,    
  sendP pipe TheEnd

 `catch` (\e -> do programTerminated =: True; throwIO e)   --    ,   c_compress()      



-- |    .

--   ,    
read_file command (startFile, _, _, _) _ file  | fi<-cfFileInfo file, fiIsDir fi = do
  startFile fi
  return$ Just$ makeCompressedFile 0 fi

--     ,     ,     
read_file _ (startFile, correctTotals, receiveBuf, sendBuf) _ (DiskFile old_fi) = do
  --     -             :)
  whenJustM (tryOpen$ diskName old_fi)   $ \file -> do
  --      ,    
  whenJustM (rereadFileInfo old_fi file) $ \fi -> do
  correctTotals$  fiSize fi - fiSize old_fi  --   UI,     
  startFile fi                               --      
  crc       <- ref aINIT_CRC                 --   CRC
  bytesRead <- ref 0                         --                 
  repeat_untilM $ do --    ,     :
    (buf, size) <- receiveBuf                --      
    len         <- fileReadBuf file buf size --        
    bytesRead   += i len                     --    
    crc        .<- updateCRC buf len         --  CRC  
    sendBuf        buf size len              --    
    return (len==0)                          --   ,   
  fileClose file
  crc'        <-  val crc >>== finishCRC     --    CRC
  bytesRead'  <-  val bytesRead              --  
  correctTotals$  bytesRead' - fiSize fi     --   UI,       getFileInfo
  return$ Just$ makeCompressedFile crc' fi{fiSize=bytesRead'}

--       ,   ,     
read_file _ (startFile, _, receiveBuf, sendBuf) decompress_pipe compressed_file = do
  crc  <-  ref aINIT_CRC                       --   CRC
  --  ""        
  --       
  let writer inbuf 0 = send_backP decompress_pipe ()  --  ,    
      writer inbuf insize = do
        (buf, size) <- receiveBuf              --      
        let len  = min insize size             --      
        crc    .<- updateCRC inbuf len         --  CRC  
        copyBytes  buf inbuf len               --     
        sendBuf    buf size len                --      
        writer     (inbuf+:len) (insize-len)   --   ,  
  let fi  =  cfFileInfo compressed_file
  startFile fi
  decompress_file decompress_pipe compressed_file writer   --     
  crc'  <-  val crc >>== finishCRC            --    CRC
  if cfCRC compressed_file == crc'            --  CRC  
    then return$ Just compressed_file         --       
    else registerError$ BAD_CRC$ diskName fi  --      



---------------------------------------------------------------------------------------------------
----       ------------------------------------------------------------
---------------------------------------------------------------------------------------------------

-- |      ,      .
--    backdoor    ,    
compress_AND_write_to_archive_PROCESS archive backdoor pipe = do

  --    UI  
  let display (FileStart fi)         =  uiStartFile      fi
      display (DataChunk buf len)    =  uiUnpackedBytes  (i len)
      display (CorrectTotals addme)  =  uiCorrectTotal   addme
      display (FakeFiles cfiles)     =  uiFakeFiles      (map cfFileInfo cfiles) 0
      display _                      =  return ()

  --      
  let write_to_archive (DataBuf buf len) =  do uiCompressedBytes  (i len)
                                               archiveWriteBuf    archive buf len
                                               return len
      write_to_archive  NoMoreData       =  return 0

  --    -       
  let copy_block = do
        CopySolidBlock files <- receiveP pipe
        let block       = (cfArcBlock (head files))
        uiFakeFiles       (map cfFileInfo files)  (blCompSize block)
        archiveCopyData   (blArchive block) (blPos block) (blCompSize block) archive
        DataEnd <- receiveP pipe
        return ()

  repeat_whileM (receiveP pipe) (notTheEnd) $
    \(CompressData block_type compressor just_copy) -> do
        case block_type of             --  UI      
          DATA_BLOCK  ->  uiStartFiles
          DIR_BLOCK   ->  uiStartDirectory
          _           ->  uiStartControlData
        result <- ref 0   --  ,     write_to_archive

        --  CRC (   )       
        crc      <- ref aINIT_CRC
        origsize <- ref 0
        let update_crc (DataChunk buf len) =  do when (block_type/=DATA_BLOCK) $ do
                                                   crc .<- updateCRC buf len
                                                 origsize += i len
            update_crc _                   =  return ()

        let compressP  =  de_compress_PROCESS freearcCompress   --    
        --   ,    `compressor`
        --let compressa  =  foldl (|>) storing_PROCESS (map compressP compressor)
        let compressa = case compressor
              of []        -> storing_PROCESS
                 [a]       -> storing_PROCESS |> compressP a
                 [a,b]     -> storing_PROCESS |> compressP a |> compressP b
                 [a,b,c]   -> storing_PROCESS |> compressP a |> compressP b |> compressP c
                 [a,b,c,d] -> storing_PROCESS |> compressP a |> compressP b |> compressP c |> compressP d
                 _         -> \pipe -> fail "More than 4 compressors in line currently is not supported"
        --  ,         / 
        let compress_block  =  runFuncP compressa (do x<-receiveP pipe; display x; update_crc x; return x)
                                                  (send_backP pipe)
                                                  (write_to_archive .>>= writeIORef result)
                                                  (val result)
        --         -   
        let compress_f  =  if just_copy  then copy_block  else compress_block

        --   -
        pos_begin <- archiveGetPos archive
        ; uiStartDeCompression       --    
        ;   compress_f               --  
        ; uiFinishDeCompression      --     (/            )
        ; uiUpdateProgressIndicator  -- ,     
        pos_end   <- archiveGetPos archive

        --          
        --       
        (Directory dir)  <-  receiveP pipe   --        
        crc'             <-  val crc >>== finishCRC     --    CRC
        origsize'        <-  val origsize
        putP backdoor (ArchiveBlock {
                           blArchive    = archive
                         , blType       = block_type
                         , blCompressor = compressor
                         , blPos        = pos_begin
                         , blOrigSize   = origsize'
                         , blCompSize   = pos_end-pos_begin
                         , blCRC        = crc'
                         , blFiles      = undefined
                       }, dir)


-- | ,   Instruction   CompressionData
storing_PROCESS pipe = do
  let send (DataChunk buf len)  =  resend_data pipe (DataBuf buf len)  >>  send_backP pipe (buf,len)
      send  DataEnd             =  resend_data pipe NoMoreData >> return ()
      send _                    =  return ()

  --   
  repeat_whileM (receiveP pipe) (notDataEnd) (send)
  send DataEnd   --   ,    
  return ()


{-# NOINLINE writeBlock #-}
{-# NOINLINE read_file #-}
{-# NOINLINE storing_PROCESS #-}

