----------------------------------------------------------------------------------------------------
----      ( "l")    ( "v").     ----
----------------------------------------------------------------------------------------------------
module ArcList (runArchiveList, runCommentWrite) where

import Prelude hiding (catch)
import Control.Concurrent
import Control.Exception
import Control.Monad
import Data.IORef
import Foreign.C.String
import Foreign.C.Types
import System.IO
import Numeric

import Process
import Utils
import Errors
import Files
import FileInfo
import Cmdline
import CUI
import ArhiveStructure


-- |    
runArchiveList command@Command{ cmd_arclist     = arclist
                              , cmd_arcname     = arcname
                              , opt_arc_basedir = arc_basedir
                              , cmd_arc_filter  = arc_filter
                              } = do
  archive <- archiveReadInfo command arc_basedir "" arc_filter arcname
  results <- archiveList command (length arclist==1) archive
  arcClose archive
  return results

-- | 
archiveList command @ Command{ cmd_name = cmd, cmd_arcname = arcname }
            show_empty
            ArchiveInfo { arcDirectory = directory
                        , arcComment   = arcComment
                        } = do
  let files = length directory
      bytes = sum$ map (fiSize.cfFileInfo) directory
  when (files>0 || show_empty) $ do
    uiStartArchive command ""  --       
    uiPrintArcComment arcComment
    case cmd of
      "l" -> do putStrLn$ "Date/time                  Size Filename"
                putStrLn$ "----------------------------------------"
                mapM_ list directory
                putStrLn$ "----------------------------------------"
                putStr$   show3 files ++ " files, " ++ show3 bytes ++ " bytes"

      "v" -> do putStrLn$ "Date/time              Attr            Size          Packed      CRC Filename"
                putStrLn$ "-----------------------------------------------------------------------------"
                mapM_ verbose_list directory
                putStrLn$ "-----------------------------------------------------------------------------"
                putStr$   show3 files ++ " files, " ++ show3 bytes ++ " bytes"

  return (1, files, bytes, -1)

-- |   
list direntry = do
  let fi = (cfFileInfo direntry)
  timestr <- formatDateTime (fiTime fi)
  putStrLn$        timestr
         ++ " " ++ right_justify 11 (if (fiIsDir fi) then ("-dir-") else (show3$ fiSize fi))
         ++ " " ++ (str2terminal$ fpFullname$ fiStoredName fi)

-- |   
verbose_list direntry = do
  let fi = (cfFileInfo direntry)
  timestr <- formatDateTime (fiTime fi)
  putStrLn$        timestr
         ++ " " ++ (if (fiIsDir fi)  then ".D....."  else ".......")
         ++ " " ++ right_justify 15 (show$ fiSize fi)
         ++ " " ++ right_justify 15 "0"
         ++ " " ++ left_fill  '0' 8 (showHex (cfCRC direntry) "")
         ++ " " ++ (str2terminal$ fpFullname$ fiStoredName fi)

{-
-- |   
technical_list direntry = do
  let fi = (cfFileInfo direntry)
  timestr <- formatDateTime (fiTime fi)
  putStrLn$ ""
  putStrLn$ "Filename: "  ++ (str2terminal$ fpFullname$ fiStoredName fi)
  putStrLn$ "Size: "      ++ (show$ fiSize fi)
  putStrLn$ "Date/time: " ++ timestr
  putStrLn$ "CRC: "       ++ showHex (cfCRC direntry) ""
  putStrLn$ "Type: "      ++ if (fiIsDir fi) then "directory" else "file"
-}

-- | CTime     "%Y-%m-%d %H:%M:%S"
formatDateTime t  =  c_FormatDateTime t >>= peekCString

foreign import ccall unsafe "Environment.h FormatDateTime"
  c_FormatDateTime :: CTime -> IO CString

{-
mapMMM action timeformat directory = do
  --let format  =  showtime timeformat . convert_CTime_to_ClockTime
  let format t =  c_FormatDateTime t >>= peekCString
  timestr <- format aMINIMAL_POSSIBLE_DATETIME
  ref_time <- newIORef (aMINIMAL_POSSIBLE_DATETIME, timestr)
  cnt <- newIORef 0
  let new_action direntry = do
        let fi  =  cfFileInfo direntry
            newtime  =  fiTime fi
        modifyIORefIO ref_time $ \(time,timestr) -> do
          --modifyIORef cnt (if time /= newtime  then (1+)  else id)
          --new_timestr  <-  if time /= newtime  then format newtime  else return timestr
          new_timestr  <-  format newtime
          action direntry fi new_timestr
          return (newtime, new_timestr)
  mapM_ new_action directory
  print =<< readIORef cnt
-}


-- |     
runCommentWrite command@Command{ cmd_filespecs   = filespecs
                               , cmd_arcname     = arcname
                               } = do
  when (length filespecs /= 1) $
    registerError$ GENERAL_ERROR$ "command syntax is \"cw archive outfile\""
  let [outfile] = filespecs
  putStr$ "Writing archive comment of "++arcname++" to "++outfile
  bracket (archiveReadInfo command "" "" (const False) arcname)
          (arcClose)
          (\archive -> filePut outfile (arcComment archive))
  return (0,0,0,0)
