----------------------------------------------------------------------------------------------------
---- , ,        .    ------
----     :                                                      ------
----   *   ,                                            ------
----   *     ,  -                            ------
----   *                             ------
----------------------------------------------------------------------------------------------------
module ArhiveFileList where

import Data.HashTable as Hash
import Data.List
import Data.Maybe

import Utils
import Files
import FileInfo
import Cmdline
import ArhiveStructure

----------------------------------------------------------------------------------------------------
----           -------------------------
----------------------------------------------------------------------------------------------------

-- |Sort `filelist` according to order given in `command`
sort_files Command{ opt_sort_order    = sort_order     --     ,  "gen" -   ,   
                  , opt_find_group    = find_group     --  "FileInfo ->  "
                  , opt_groups_count  = groups_count   --   (`find_group`     0..groups_count-1)
                  }
           filelist = dirs ++ sortBy sort_order files  --   ,     
  where
    (dirs,files)  =  partition fiIsDir filelist
    sortBy sortOrder =
      case sortOrder of
        ""     -> id
        'g':xs -> concatMap (sortBy xs) . partitionList groups_count find_group   --           
        "en"   -> sortOn  en_key
        "es"   -> sortOn' es_key
        _      -> sortOn  (key_func sortOrder find_group)

    en_key fi  =  (fpLCExtension$ fiStoredName fi, filenameLower$ fpBasename$ fiStoredName fi)
    es_key fi  =  (fpLCExtension$ fiStoredName fi, fiSize fi)
{-
    --     (-ogei)
    intellectual_sort sortOrder =
    -- 1.     "ge"
    -- 2.          
    -- 3a.  ,      -      "s" ("ps",        ?)
    -- 3b.      "ns"
    -- 4.       ?     ?
      let (o1,'i':o2)  =  break (=='i') sortOrder
      in concatMap isort . sort_and_groupOn' (key_func o1 find_group)
      where isort group =      --   ,   
              let groups = groupOn three (sortOn ns_key group)  --    3-  
                             where three = take 3 . filenameLower.fpBasename.fiStoredName
                  (singles, full_groups)  =  partition (null.tail) groups  --   ,   - ,  " " :)
                  list1  =  sortOn' s_key (concat singles)  --      
              in list1 ++ concat full_groups

            ns_key fi  =  (filenameLower$ fpBasename$ fiStoredName fi, fiSize fi)
            np_key fi  =  (filenameLower$ fpBasename$ fiStoredName fi, fpDirectory$ fiStoredName fi)
            s_key  fi  =  fiSize fi
            e_key  fi  =  fpLCExtension$ fiStoredName fi
            n_key  fi  =  filenameLower$ fpBasename$ fiStoredName fi
-}

-- |Map `sort_order` to function returning ordering key
key_func sort_order find_group  =  map_functions (map key sort_order)
  where key 'p' = OrderFilePath .filenameLower.fpDirectory  .fiStoredName
        key 'n' = OrderFilePath .filenameLower.fpBasename   .fiStoredName
        key 'e' = OrderFilePath               .fpLCExtension.fiStoredName
        key 's' = OrderFileSize                             .fiSize
        key 't' = OrderFileTime                             .fiTime
        key 'g' = OrderGroup                                .find_group

-- |     
data SortOrder =   OrderFilePath  !FilePath
                 | OrderPackedStr !MyPackedString
                 | OrderFileSize  !FileSize
                 | OrderFileTime  !FileTime
                 | OrderGroup     Int               deriving (Eq, Ord)



----------------------------------------------------------------------------------------------------
----           -----------------------
----------------------------------------------------------------------------------------------------

-- |          (  /)
--            ,       ,
--   -   ,          (          )
splitBy []     =  (:[])  --       -     
splitBy crits  =  splitByLen (computeLen)  where
  computeLen xs = len  where
    minlen   = minimum$ map (`splitLenMin` xs) crits   --    
    maxlen   = minimum$ map (`splitLenMax` xs) crits   --    
    solidLen = solidBlockLen xs                        --  -   
    len      = if minlen<=solidLen && solidLen<=maxlen
                 then solidLen  --  -   ,   
                 else minlen    --         
    --to do: GroupByBlockSize    
    --to do:   16 , - 40  ->      20  ?

-- |   (   ),   
splitLenMin  GroupNone              = const 1
splitLenMin  GroupByExt             = length.head.groupOn (fpLCExtension.fiFilteredName.cfFileInfo)
splitLenMin (GroupBySize size)      = (1+)      . groupLen (fiSize.cfFileInfo) (+) (<i size)
splitLenMin (GroupByBlockSize size) = atLeast 1 . groupLen (fiSize.cfFileInfo) (+) (<i size)
splitLenMin (GroupByNumber n)       = const n
splitLenMin  GroupAll               = const maxBound

-- |   ,    
-- (            ,
--    )
splitLenMax (GroupByNumber n)       = const (n*2)
splitLenMax (GroupBySize size)      = splitLenMin (GroupBySize (size*2))
splitLenMax x                       = splitLenMin x



----------------------------------------------------------------------------------------------------
----          -----------------------------
----------------------------------------------------------------------------------------------------

-- |  ,      ,
--      ,     
--       .
--         `update_type` (a/f/u/s)
--
join_lists main_archive added_archives added_diskfiles
           Command {            --    :
               opt_update_type = update_type    --    (a/f/u/s)
             , opt_append      = append         --       ?
             , opt_sort_order  = sort_order     --     ,  "gen" -   ,   
             , opt_find_group  = find_group     --  "FileInfo ->  "
             } = do

  --  :
  -- 1.   ,        , .
  --      -     ( StoredName)  
  -- 2.      -   
  -- 3.      ,      ( ),
  --          ,      
  -- 4.              ,
  --         

  let --     ,   
      main_list   =  arcDirectory main_archive
      --     ,          
      added_list  =  (concatMap arcDirectory added_archives)  ++  (map DiskFile added_diskfiles)

  -- DEBUGGING ------------------------------------------------------
  let typ (DiskFile _)             = "Disk: "
      typ (CompressedFile _ _ _ _) = "Archive: "
      name  fi = fpFullname(fiStoredName fi)
      names fi = fpFullname(fiDiskName fi)++" "++fpFullname(fiFilteredName fi)++" "++fpFullname(fiStoredName fi)
  --print$ map (names.cfFileInfo) arcdir
  --print$ map (names.cfFileInfo) diskfiles
  -- DEBUGGING ------------------------------------------------------

  -- 1.   :     ,    ,
  --           
  let keyFunc  =  fiStoredName . cfFileInfo  -- ,     -
  (main_list,  main_hash )  <-  removeDuplicates main_list  keyFunc (==) fpHash
  (added_list, added_hash)  <-  removeDuplicates added_list keyFunc (==) fpHash

  -- 2.      -   
  case () of
   _ | null main_list   -> return added_list  --     ,    
     | null added_list  -> return (if update_type=='s' then [] else main_list)  --  
     | otherwise        -> do
    --     :

    -- 3.         ,  ,  
    let newer_file arcfile diskfile =   --   ,  
          if (fiTime (cfFileInfo arcfile) >= fiTime (cfFileInfo diskfile))
            then arcfile   --       ,   
            else diskfile  --    ,   

    let sync_file arcfile diskfile =   --   ,      
          if (fiTime (cfFileInfo arcfile) == fiTime (cfFileInfo diskfile))
            then arcfile   --     ,   
            else diskfile  --      ,   

    let select_file arcfile = do
          diskfile <- Hash.lookup added_hash (keyFunc arcfile)
          case (diskfile, update_type) of
            (Nothing, 's')        ->  return$ Nothing        --       :  "--sync"    ,    
            (Nothing,  _ )        ->  return$ Just arcfile   --          
            (Just diskfile, 'a')  ->  return$ Just diskfile  --     ,   :   "a"     
            (Just diskfile, 's')  ->  return$ Just (sync_file  arcfile diskfile)  --  "--sync"   ,        
            (Just diskfile,  _ )  ->  return$ Just (newer_file arcfile diskfile)  --        

    list1 <- mapMaybeM select_file main_list  --          


    -- 4.      ,       
    let new_files_only diskfile = do
          --  ,        
          isNothing ==<< Hash.lookup main_hash (keyFunc diskfile)

    list2 <- case (update_type) of
                 --  "f": ,    ,  
               'f' -> return []
                 --     ,      
               _   -> Utils.filterM new_files_only added_list

    -- DEBUGGING ------------------------------------------------------
    --print$ map (\f -> typ f ++ (fpFullname.fiStoredName.cfFileInfo) f) list1
    --print$ map (fpFullname.fiStoredName.cfFileInfo) list2
    -- DEBUGGING ------------------------------------------------------

    --            
    return$ joinFilelists append sort_order find_group list1 list2


-- |  (   +   )
joinFilelists True  _          _           =  (++)      --    "--append"
joinFilelists False ""         _           =  (++)      --     ("-ds")
joinFilelists False sort_order find_group  =  merge (map2cmp (key_func sort_order find_group . cfFileInfo))


-- |    `originalList` -  ,
--      
--
removeDuplicates originalList keyFunc eqFunc hashFunc = do
  table <- Hash.new eqFunc hashFunc

  --    `table`  `value` ,          .
  --         ,    `update`
  -- (to do: `reverse`   `update`)
  let insert_no_dups value = do
        let key  =  keyFunc value
        found <- Hash.lookup table key
        case (found) of
          Nothing       ->  do Hash.insert table key value
                               return True
          Just oldfile  ->  return False

  list <- Utils.filterM insert_no_dups originalList
  return (list,table)

{-# NOINLINE splitBy #-}
{-# NOINLINE removeDuplicates #-}
