#ifndef FREEARC_COMMON_H
#define FREEARC_COMMON_H

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

/******************************************************************************
**   ****************************************************
******************************************************************************/
#define make4byte(a,b,c,d)       ((a)+256*((b)+256*((c)+256*(((uint32)d)))))
#define strequ(a,b)              (strcmp((a),(b))==0)
#define FreeAndNil(ptr)          {delete (ptr); (ptr)=NULL;}
#define iterate(num, statement)  {for( int i=0; i<num; i++) {statement;}}
#define iterate_var(i, num)      for( int i=0; i<num; i++)
#define iterate_array(i, array)  for( int i=0; i<array.size; i++)
#define TRUE                     1
#define FALSE                    0

#define nameequ(s1,s2)           (!namecmp(s1,s2))
#define in_set( c, set )         (strchr (set, c ) != NULL)
#define in_set0( c, set )        (memchr (set, c, sizeof(set) ) != 0)
#define str_end(str)             (strchr (str,'\0'))
#define last_char(str)           (str_end(str) [-1])
#define strdup_msg(s)            (strcpy (new char[strlen(s)+1], (s)))
#define find_extension(str)      (find_extension_in_entry (parse_path(str)))
#define mymax(a,b)               ((a)>(b)? (a) : (b))
#define mymin(a,b)               ((a)<(b)? (a) : (b))
#define char2int(c)              ((c)-'0')
#define elements(arr)            (sizeof(arr)/sizeof(*arr))
#define endof(arr)               ((arr)+elements(arr))

#define file_exists(name)        (access((name),0) == 0)

#ifndef CHECK
#define CHECK(a,b)               {if (!(a))  printf b, exit(1);}
#endif

/******************************************************************************
**    ,    ***********************
******************************************************************************/
typedef unsigned int      uint;
typedef unsigned char     uint8;
typedef unsigned short    uint16;
typedef unsigned int      uint32;
typedef unsigned __int64  uint64;

typedef char*    FILENAME;         //  
#ifdef FREEARC_LARGE_FILES
typedef __int64  FILESIZE;         //  
#else
typedef uint     FILESIZE;
#endif


#ifdef __cplusplus
/******************************************************************************
** ,  ,    :) *************************
******************************************************************************/
template <typename T> class ARRAY
{
public:
  int size;                                                  // -   
  T  *data;                                                  // ,   
  bool autodelete;                                           //   data    

  void setsize (int _size)          {size = _size; data = size? new T[size] : NULL; autodelete=TRUE;}
  void resize (int _size)           {if(autodelete) delete data; setsize(_size);}   //     
  void set (int _size, void* ptr)   {resize(0); size=_size, data=(T*)ptr, autodelete=FALSE;}  //         
  ARRAY (int _size=0)               {setsize (_size);}       //     _size
  ~ARRAY()                          {resize(0);}
  T& operator[] (int i)             {return data[i];}
  T& operator() (int i)             {return data[i];}
};


/******************************************************************************
** ,     *************************************
******************************************************************************/
enum MODE {READ_MODE, WRITE_MODE}; //   
class MYFILE
{
  int handle;
public:
  MYFILE& open (FILENAME filename, MODE mode)    //      
  {
    handle = ::open (filename, mode==READ_MODE? O_RDONLY|O_BINARY : O_WRONLY|O_BINARY|O_CREAT|O_TRUNC, S_IREAD|S_IWRITE);
    CHECK (handle>=0, ("\nERROR: can't open file %s", filename));
  }
  void close()    //  
  {
    CHECK (::close(handle)==0, ("\nERROR: can't close file"));
    handle = -1;
  }

#ifdef FREEARC_LARGE_FILES
  FILESIZE size    ()                {return _filelengthi64 (handle);}                         //   
  void     seek    (FILESIZE pos)    {CHECK( _lseeki64 (handle, SEEK_SET, pos) == pos, ("\nERROR: file seek operation failed"));}       //      
#else
  FILESIZE size    ()                {return filelength (handle);}
  void     seek    (FILESIZE pos)    {CHECK( lseek (handle, pos, SEEK_SET) == pos, ("\nERROR: file seek operation failed"));}
#endif

  FILESIZE tryRead (void *buf, FILESIZE size)   {int result = ::read (handle, buf, size); CHECK(result>=0, ("\nERROR: file read operation failed")); return result;}           //  -  ,     
  void     read    (void *buf, FILESIZE size)   {CHECK (tryRead (buf, size) == size, ("\nERROR: can't read %lu bytes", (unsigned long)size));}         //  ,       
  void     write   (void *buf, FILESIZE size)   {CHECK (::write (handle, buf, size) == size, ("\nERROR: file write operation failed"));}

  MYFILE ()                               {handle=-1;}
  MYFILE (FILENAME filename, MODE mode)   {open (filename, mode);}
  ~MYFILE()                               {if (handle>=0) close();}
};

#endif  // __cplusplus
#endif
