#ifndef FREEARC_COMPRESSION_H
#define FREEARC_COMPRESSION_H

// 
#define FREEARC_OK                               0     /* ALL RIGHT */
#define FREEARC_ERRCODE_GENERAL                  (-1)  /* Some error when (de)compressing */
#define FREEARC_ERRCODE_INVALID_COMPRESSOR       (-2)  /* Invalid compression method or parameters */
#define FREEARC_ERRCODE_ONLY_DECOMPRESS          (-3)  /* Program builded with FREEARC_DECOMPRESS_ONLY, so don't try to use compress */
#define FREEARC_ERRCODE_OUTBLOCK_TOO_SMALL       (-4)  /* Output block size in (de)compressMem is not enough for all output data */
#define FREEARC_ERRCODE_NOT_ENOUGH_MEMORY        (-5)  /* Can't allocate memory needed for (de)compression */

//      
#define b_ 1
#define kb 1024
#define mb (kb*kb)
#define gb (kb*mb)

//  ,   /      
#define BUFFER_SIZE (64*kb)

//  ,     LZMA Decode (  ;)
#define LZ_CHUNKS (256*kb)

//         
#define COMPRESSION_METHOD_PARAMETERS_DELIMITER     ':'   /*        */
#define MAX_COMPRESSION_METHODS  20          /*       ,    AddCompressionMethod */
#define MAX_PARAMETERS           20          /*      -  ( ),      */
#define MAX_METHOD_STRLEN        256         /*   ,    */
#define mymin(a,b)               ((a)<(b)? (a) : (b))
#define mymax(a,b)               ((a)>(b)? (a) : (b))


#ifdef __cplusplus
extern "C" {
#endif

//          /
typedef int INOUT_FUNC (void *buf, int size);

//  ,   
int decompress (char *method, INOUT_FUNC *read_f, INOUT_FUNC *write_f);
//            
int DecompressWithHeader (INOUT_FUNC *read_f, INOUT_FUNC *write_f);
//    ,       outputSize .
//      ,    
int DecompressMem (char *method, void *input, int inputSize, void *output, int outputSize);
int DecompressMemWithHeader     (void *input, int inputSize, void *output, int outputSize);

#ifndef FREEARC_DECOMPRESS_ONLY
//    
int compress   (char *method, INOUT_FUNC *read_f, INOUT_FUNC *write_f);
//            
int CompressWithHeader (char *method, INOUT_FUNC *read_f, INOUT_FUNC *write_f);
//    ,       outputSize .
//      ,    
int CompressMem           (char *method, void *input, int inputSize, void *output, int outputSize);
int CompressMemWithHeader (char *method, void *input, int inputSize, void *output, int outputSize);
//   out_method     in_method ( ParseCompressionMethod + ShowCompressionMethod)
int CanonizeCompressionMethod (char *in_method, char *out_method);
//   ,   /,     .
int GetCompressionMem   (char *method);
int GetDecompressionMem (char *method);
int GetDictionary       (char *method);
int GetBlockSize        (char *method);
//   out_method   ,   
//   // 
int SetCompressionMem   (char *in_method, int mem,  char *out_method);
int SetDecompressionMem (char *in_method, int mem,  char *out_method);
int SetDictionary       (char *in_method, int dict, char *out_method);
int SetBlockSize        (char *in_method, int bs,   char *out_method);
//   out_method   , ,  ,
//    /   /  
int LimitCompressionMem   (char *in_method, int mem,  char *out_method);
int LimitDecompressionMem (char *in_method, int mem,  char *out_method);
int LimitDictionary       (char *in_method, int dict, char *out_method);
int LimitBlockSize        (char *in_method, int bs,   char *out_method);
#endif

//  "()",     
int copy_data   (INOUT_FUNC *read_f, INOUT_FUNC *write_f);



#ifdef __cplusplus

//      
class COMPRESSION_METHOD
{
public:
  //    
  virtual int decompress (INOUT_FUNC *read_f, INOUT_FUNC *write_f) = 0;
#ifndef FREEARC_DECOMPRESS_ONLY
  virtual int compress   (INOUT_FUNC *read_f, INOUT_FUNC *write_f) = 0;

  //   buf[MAX_METHOD_STRLEN] ,       (,   ParseCompressionMethod)
  virtual void ShowCompressionMethod (char *buf) = 0;

  //   ,   /,
  //   (          -  lz/bs ),
  //    (          - -  bs   lzp)
  virtual int  GetCompressionMem   (void)     = 0;
  virtual int  GetDecompressionMem (void)     = 0;
  virtual int  GetDictionary       (void)     = 0;
  virtual int  GetBlockSize        (void)     = 0;
  //       - ,    
  virtual void SetCompressionMem   (int mem)  = 0;
  virtual void SetDecompressionMem (int mem)  = 0;
  virtual void SetDictionary       (int dict) = 0;
  virtual void SetBlockSize        (int bs)   = 0;
  //    / ,   /  
  void LimitCompressionMem   (int mem)  {if (GetCompressionMem()  > mem)  SetCompressionMem(mem);}
  void LimitDecompressionMem (int mem)  {if (GetDecompressionMem()> mem)  SetDecompressionMem(mem);}
  void LimitDictionary       (int dict) {if (GetDictionary()      > dict) SetDictionary(dict);}
  void LimitBlockSize        (int bs)   {if (GetBlockSize()       > bs)   SetBlockSize(bs);}
#endif
};

//    -  COMPRESSION_METHOD,
//   ,     `method`
COMPRESSION_METHOD *ParseCompressionMethod (char* method);


typedef COMPRESSION_METHOD* (*CM_PARSER) (char** parameters);
int AddCompressionMethod (CM_PARSER parser);  //         
int parseInt (char *param, int *error);  //   param    -  ,   error=1
int parseMem (char *param, int *error);  // ,   param    b/k/m/g/^,      (  - '^', ..  )
void showMem (int mem, char *result);    //     

void strncopy( char *to, char *from, int len );
void split (char *str, char splitter, char **result);


//   "" STORING
class STORING_METHOD : public COMPRESSION_METHOD
{
public:
  //    
  virtual int decompress (INOUT_FUNC *read_f, INOUT_FUNC *write_f);
#ifndef FREEARC_DECOMPRESS_ONLY
  virtual int compress   (INOUT_FUNC *read_f, INOUT_FUNC *write_f);

  //   buf[MAX_METHOD_STRLEN] ,    (,   parse_STORING)
  virtual void ShowCompressionMethod (char *buf);

  // /  ,   /,     
  virtual int  GetCompressionMem     (void) {return BUFFER_SIZE;}
  virtual int  GetDecompressionMem   (void) {return BUFFER_SIZE;}
  virtual int  GetDictionary         (void) {return 0;}
  virtual int  GetBlockSize          (void) {return 0;}
  virtual void SetCompressionMem     (int)  {}
  virtual void SetDecompressionMem   (int)  {}
  virtual void SetDictionary         (int)  {}
  virtual void SetBlockSize          (int)  {}
#endif
};

//     STORING
COMPRESSION_METHOD* parse_STORING (char** parameters);

}       // extern "C"
#endif  // __cplusplus

#endif
