/*
 *  Example program of using Compression-2005 library with C/C++
 *
 *  This program implements simple file-to-file compressor and decompressor
 *  Run program without parameters to see it's syntax
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>

#include "Compression.h"

// Input/output files and callbacks for (de)compression functions
FILE *infile, *outfile;
int read_func (void *buf, int size)
{
  return fread(buf,1,size,infile);
}
int write_func (void *buf, int size)
{
  return fwrite(buf,1,size,outfile)==size? size : -1;
}

int main(int argc, char *argv[])
{
  if (argc==4)
  {
    infile  = fopen (argv[2], "rb");   if (infile==NULL)  {printf ("Can't open input file %s!\n", argv[2]); return EXIT_FAILURE;}
    outfile = fopen (argv[3], "wb");   if (outfile==NULL) {printf ("Can't create output file %s!\n", argv[3]); return EXIT_FAILURE;}
    int result;
    if (strcmp (argv[1], "d"))
      result = CompressWithHeader (argv[1], read_func, write_func);
    else
      result = DecompressWithHeader (read_func, write_func);
    fclose (infile);
    fclose (outfile);
    if (result) {printf ("(De)compression error %d!\n", result); return EXIT_FAILURE;}
    return EXIT_SUCCESS;
  }
  else
  {
    puts("Usage, compression: Compressor method infile outfile\n"
         "       decompression: Compressor d infile outfile\n"
         "Supported methods are \"ppmd:...\", \"lzma:...\", \"grzip:...\"");
    return argc==1? EXIT_SUCCESS : EXIT_FAILURE;
  }
}
