// C_LZMA.cpp -  FreeArc    LZMA

#ifdef WIN32
#include <windows.h>
#include <initguid.h>
#else
#define INITGUID
#endif

// #include <limits.h>
#include <math.h>
#include <stdio.h>
#include <basetsd.h>
typedef SSIZE_T ssize_t;

#if defined(WIN32) || defined(OS2) || defined(MSDOS)
#include <fcntl.h>
#include <io.h>
#define MY_SET_BINARY_MODE(file) setmode(fileno(file),O_BINARY)
#else
#define MY_SET_BINARY_MODE(file)
#endif

extern "C" {
#include "C_LZMA.h"
}

//    .o    
#include "Common/Alloc.cpp"
#include "7zip/Common/InBuffer.cpp"
#include "7zip/Compress/LZ/LZOutWindow.cpp"
#include "7zip/Compress/LZMA/LZMADecoder.cpp"
#ifndef FREEARC_DECOMPRESS_ONLY
#include "Common/CRC.cpp"
#include "7zip/Common/OutBuffer.cpp"
#include "7zip/Compress/LZ/LZInWindow.cpp"
#include "7zip/Compress/RangeCoder/RangeCoderBit.cpp"
#include "7zip/Compress/LZMA/LZMAEncoder.cpp"
#endif  // !defined (FREEARC_DECOMPRESS_ONLY)


class CallbackInStream:
  public ISequentialInStream,
  public CMyUnknownImp
{
  INOUT_FUNC *read_f;
public:
  int failed;
  MY_UNKNOWN_IMP

  CallbackInStream(INOUT_FUNC *_read_f) : read_f(_read_f), failed(0) {}
  virtual ~CallbackInStream() {}
  STDMETHOD(Read)(void *data, UInt32 size, UInt32 *processedSize);
  STDMETHOD(ReadPart)(void *data, UInt32 size, UInt32 *processedSize);
};

STDMETHODIMP CallbackInStream::Read(void *data, UInt32 size, UInt32 *processedSize)
{
  if(processedSize != NULL)
    *processedSize = 0;
  ssize_t res = read_f( data, mymin( (size_t) size, BUFFER_SIZE));
  if (res < 0) {
    failed = 1;
    return E_FAIL;
  }
  if(processedSize != NULL)
    *processedSize = (UInt32)res;
  return S_OK;
}

STDMETHODIMP CallbackInStream::ReadPart(void *data, UInt32 size, UInt32 *processedSize)
{
  return Read(data, size, processedSize);
}



class CallbackOutStream:
  public ISequentialOutStream,
  public CMyUnknownImp
{
  INOUT_FUNC *write_f;
public:
  int failed;
  MY_UNKNOWN_IMP

  CallbackOutStream(INOUT_FUNC *_write_f) : write_f(_write_f), failed(0) {}
  virtual ~CallbackOutStream() {}
  STDMETHOD(Write)(const void *data, UInt32 size, UInt32 *processedSize);
  STDMETHOD(WritePart)(const void *data, UInt32 size, UInt32 *processedSize);
};

STDMETHODIMP CallbackOutStream::Write(const void *data, UInt32 size, UInt32 *processedSize)
{
  if(processedSize != NULL)
    *processedSize = 0;
  ssize_t res = write_f((void*)data, (size_t)size);
  if (res < 0) {
    failed = 1;
    return E_FAIL;
  }
  if(processedSize != NULL)
    *processedSize = (UInt32)res;
  return S_OK;
}

STDMETHODIMP CallbackOutStream::WritePart(const void *data, UInt32 size, UInt32 *processedSize)
{
  return Write(data, size, processedSize);
}


#ifndef FREEARC_DECOMPRESS_ONLY

int lzma_compress  ( int dictionarySize,
                     int algorithm,
                     int numFastBytes,
                     int matchFinder,
                     int posStateBits,
                     int litContextBits,
                     int litPosBits,
                     INOUT_FUNC *read_f,
                     INOUT_FUNC *write_f)
{
  CallbackInStream   inStream  (read_f);
  CallbackOutStream  outStream (write_f);
  NCompress::NLZMA::CEncoder* encoder = new NCompress::NLZMA::CEncoder;

  bool eos = (1==1);  // use End-Of-Stream marker because we don't know filesize apriori

  if (encoder->SetupProperties(dictionarySize,
                               posStateBits,
                               litContextBits,
                               litPosBits,
                               algorithm,
                               numFastBytes,
                               matchFinder,
                               eos) != S_OK)
  {
    delete encoder;
    return FREEARC_ERRCODE_INVALID_COMPRESSOR;
  }

  HRESULT result = encoder->Code(&inStream, &outStream, 0, 0, 0);
  delete encoder;
  if (inStream.failed || outStream.failed)  //       -     
    return 0;
  if (result == E_OUTOFMEMORY)
  {
    return FREEARC_ERRCODE_NOT_ENOUGH_MEMORY;
  }
  else if (result != S_OK)
  {
    fprintf(stderr, "\nEncoder error = %X\n", (unsigned int)result);
    return 1;
  }
  return 0;
}

#endif  // !defined (FREEARC_DECOMPRESS_ONLY)

int lzma_decompress( int dictionarySize,
                     int algorithm,
                     int numFastBytes,
                     int matchFinder,
                     int posStateBits,
                     int litContextBits,
                     int litPosBits,
                     INOUT_FUNC *read_f,
                     INOUT_FUNC *write_f)
{
  CallbackInStream            inStream  (read_f);
  CallbackOutStream           outStream (write_f);
  NCompress::NLZMA::CDecoder* decoder = new NCompress::NLZMA::CDecoder;

  if (decoder->SetupProperties(dictionarySize, posStateBits, litContextBits, litPosBits) != S_OK)
  {
    delete decoder;
    return 1;
  }

  UInt64 fileSize = (UInt64)-1;

  HRESULT result = decoder->Code(&inStream, &outStream, 0, &fileSize, 0);
  delete decoder;
  if (inStream.failed || outStream.failed)  //       -     
    return 0;
  if (result != S_OK)
  {
    fprintf(stderr, "\nDecoder error = %X\n", (unsigned int)result);
    return 1;
  }
  return 0;
}



/*-------------------------------------------------*/
/*   LZMA_METHOD                  */
/*-------------------------------------------------*/
#define strequ(a,b)  (strcmp((a),(b))==0)

static const char *MatchFindersTable[] =
{
  "bt2",
  "bt3",
  "bt4",
  "bt4b",
  "pat2",
  "pat2h",
  "pat3h",
  "pat4h",
  "pat2r",
  "hc3",
  "hc4"
};

//      !!!
enum MATCH_FINDERS { BT2=0, BT3, BT4, BT4B, PAT2, PAT2H, PAT3H, PAT4H, PAT2R, HC3, HC4 };

static int FindMatchFinder (const char *s)
{
  for (int m = 0; m < (int)(sizeof(MatchFindersTable) / sizeof(*MatchFindersTable)); m++)
    if (strequ(MatchFindersTable[m],s))
      return m;
  return -1;
}

//   str   start,    ,  - NULL
char* start_from (char* str, char* start)
{
  while (*start && *str==*start)  str++, start++;
  return *start? NULL : str;
}

// ,       
LZMA_METHOD::LZMA_METHOD()
{
  dictionarySize    = 8*mb;
  algorithm         = 1;
  numFastBytes      = 32;
  matchFinder       = BT4;
  posStateBits      = 2;
  litContextBits    = 3;
  litPosBits        = 0;
}

//  
int LZMA_METHOD::decompress (INOUT_FUNC *read_f, INOUT_FUNC *write_f)
{
  return lzma_decompress  (dictionarySize,
                           algorithm,
                           numFastBytes,
                           matchFinder,
                           posStateBits,
                           litContextBits,
                           litPosBits,
                           read_f,
                           write_f);
}

#ifndef FREEARC_DECOMPRESS_ONLY

//  
int LZMA_METHOD::compress (INOUT_FUNC *read_f, INOUT_FUNC *write_f)
{
  return lzma_compress  (dictionarySize,
                         algorithm,
                         numFastBytes,
                         matchFinder,
                         posStateBits,
                         litContextBits,
                         litPosBits,
                         read_f,
                         write_f);
}

//   buf[MAX_METHOD_STRLEN] ,       (,   parse_LZMA)
void LZMA_METHOD::ShowCompressionMethod (char *buf)
{
  char DictionaryStr[100], pbStr[100], lcStr[100], lpStr[100];
  showMem (dictionarySize, DictionaryStr);
  LZMA_METHOD defaults;
  sprintf (pbStr, posStateBits  !=defaults.posStateBits  ? ":pb%d" : "", posStateBits);
  sprintf (lcStr, litContextBits!=defaults.litContextBits? ":lc%d" : "", litContextBits);
  sprintf (lpStr, litPosBits    !=defaults.litPosBits    ? ":lp%d" : "", litPosBits);
  sprintf (buf, "lzma:%s:%s:%s:%d%s%s%s",
                      DictionaryStr,
                      algorithm==0? "fast": (algorithm==1? "normal":"max"),
                      MatchFindersTable [matchFinder],
                      numFastBytes,
                      pbStr,
                      lcStr,
                      lpStr);
}

// ,       
int LZMA_METHOD::GetCompressionMem (void)
{
  switch (matchFinder)
  {
    case BT2:    return dictionarySize*9 + dictionarySize/2 +  1*mb;
    case BT3:    return dictionarySize*9 + dictionarySize/2 + 65*mb;
    case BT4:    return dictionarySize*9 + dictionarySize/2 +  6*mb;
    case BT4B:   return dictionarySize*9 + dictionarySize/2 + 34*mb;
    case PAT2:   return dictionarySize* 38 +   1*mb;
    case PAT2H:  return dictionarySize* 38 +  77*mb;
    case PAT3H:  return dictionarySize* 62 +  85*mb;
    case PAT4H:  return dictionarySize*110 + 101*mb;
    case PAT2R:  return dictionarySize* 26 +   1*mb;
    case HC3:    return dictionarySize*5 + dictionarySize/2 +  1*mb;
    case HC4:    return dictionarySize*5 + dictionarySize/2 +  6*mb;
    default:     return 0;
  }
}

//    (,    )
int lb (int n)
{
  int i;
  for (i=0; n>1; i++, n/=2);
  return i;
}

//     
void LZMA_METHOD::SetCompressionMem (int mem)
{
  if (mem<=0)  return;
  switch (matchFinder)
  {
    case BT2:    dictionarySize = int(floor(mem- 1*mb)/9.5);     break;
    case BT3:    dictionarySize = int(floor(mem-65*mb)/9.5);     break;
    case BT4:    dictionarySize = int(floor(mem- 6*mb)/9.5);     break;
    case BT4B:   dictionarySize = (mem- 34*mb)/10;               break;
    case PAT2:   dictionarySize = (mem-  1*mb)/38;               break;
    case PAT2H:  dictionarySize = (mem- 77*mb)/38;               break;
    case PAT3H:  dictionarySize = (mem- 85*mb)/62;               break;
    case PAT4H:  dictionarySize = (mem-101*mb)/110;              break;
    case PAT2R:  dictionarySize = (mem-  1*mb)/26;               break;
    case HC3:    dictionarySize = int(floor(mem- 1*mb)/5.5);     break;
    case HC4:    dictionarySize = int(floor(mem- 6*mb)/5.5);     break;
  }

  //      -      
  if (dictionarySize<32*kb)
  {
    if (matchFinder!=HC3)
      matchFinder = HC3,  SetCompressionMem (mem);
    else
      dictionarySize = 32*kb;
  }

  //       2
  dictionarySize = 1 << lb(dictionarySize);
}

#endif  // !defined (FREEARC_DECOMPRESS_ONLY)

//    LZMA_METHOD    
//   NULL,          
COMPRESSION_METHOD* parse_LZMA (char** parameters)
{
  if (strcmp (parameters[0], "lzma") == 0) {
    //    ( ) - "lzma",    

    LZMA_METHOD *p = new LZMA_METHOD;
    int error = 0;  //  ,      

    //     (         )
    while (*++parameters && !error && p->matchFinder>=0)
    {
      char *param = *parameters;
           if (start_from (param, "d"))    p->dictionarySize = parseMem (param+1, &error);
      else if (start_from (param, "a"))    p->algorithm      = parseInt (param+1, &error);
      else if (start_from (param, "fb"))   p->numFastBytes   = parseInt (param+2, &error);
      else if (start_from (param, "lc"))   p->litContextBits = parseInt (param+2, &error);
      else if (start_from (param, "lp"))   p->litPosBits     = parseInt (param+2, &error);
      else if (start_from (param, "pb"))   p->posStateBits   = parseInt (param+2, &error);
      else if (start_from (param, "mf"))   p->matchFinder    = FindMatchFinder (param+2);
      else if (strequ (param, "fast"))     p->algorithm      = 0,  p->matchFinder = HC4;
      else if (strequ (param, "normal"))   p->algorithm      = 1,  p->matchFinder = BT4;
      else if (strequ (param, "max"))      p->algorithm      = 2,  p->matchFinder = BT4;
      else {
        //   ,      
        //     -  MatchFinder',   -  numFastBytes,
        //    -  dictionarySize
        int n = FindMatchFinder (param);
        if (n>=0)
          p->matchFinder = n;
        else {
          n = parseInt (param, &error);
          if (!error)  p->numFastBytes = n;
          else         error=0, p->dictionarySize = parseMem (param, &error);
        }
      }
    }
    if (error)  {delete p; return NULL;}  //     
    return p;
  } else
    return NULL;   //    lzma
}

static int LZMA_x = AddCompressionMethod (parse_LZMA);   //    LZMA
