#include <stdio.h>
#include <sys/stat.h>
#include <sys/utime.h>
#include <limits.h>
#include "Environment.h"

#ifdef FREEARC_WIN

#include <windows.h>
#include <memory.h>
#include <conio.h>
#include <time.h>

/*
void SetDateTimeAttr(const char* Filename, time_t t)
{
    struct tm* t2 = gmtime(&t);

    SYSTEMTIME t3;
    t3.wYear         = t2->tm_year+1900;
    t3.wMonth        = t2->tm_mon+1;
    t3.wDay          = t2->tm_mday;
    t3.wHour         = t2->tm_hour;
    t3.wMinute       = t2->tm_min;
    t3.wSecond       = t2->tm_sec;
    t3.wMilliseconds = 0;

    FILETIME ft;
    SystemTimeToFileTime(&t3, &ft);

    HANDLE hndl=CreateFile(Filename,GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,0);
    SetFileTime(hndl,NULL,NULL,&ft);  //creation, last access, modification times
    CloseHandle(hndl);
    //SetFileAttributes(Filename,ai.attrib);
}
*/


#define LARGE_STRING 1024

char Saved_Title[LARGE_STRING];
bool Saved = FALSE;

void EnvSetConsoleTitle (char *title)
{
  if (!Saved) {
    GetConsoleTitle (Saved_Title, sizeof(Saved_Title));
    Saved = TRUE;
  }
  SetConsoleTitle (title);
}

void EnvResetConsoleTitle (void)
{
  if (Saved) {
    SetConsoleTitle (Saved_Title);
    Saved = FALSE;
  }
}

char *ExpandProgname (char *)
{
  static char buf[LARGE_STRING];
  GetModuleFileName (NULL, buf, LARGE_STRING);
  return buf;
}

unsigned GetPhysicalMemory (void)
{
  MEMORYSTATUS buf;
  GlobalMemoryStatus (&buf);
  return buf.dwTotalPhys;
}

void SetFileDateTime(const CFILENAME Filename, time_t mtime)
{
  struct _stat st;
    _wstat( Filename, &st);
  struct _utimbuf times;
    times.actime  = st.st_atime;
    times.modtime = mtime;
  _wutime( Filename, &times);
}

#else // For Unix:

void EnvSetConsoleTitle    (char *)      {}
void EnvResetConsoleTitle  (void)        {}
char *ExpandProgname       (char *argv0) {return argv0;}
unsigned GetPhysicalMemory (void)        {return UINT_MAX;}

void SetFileDateTime(const CFILENAME Filename, time_t mtime)
{
  struct stat st;
    stat( Filename, &st);
  struct utimbuf times;
    times.actime  = st.st_atime;
    times.modtime = mtime;
  utime( Filename, &times);
}

#endif // Windows/Unix


char *FormatDateTime(time_t t)
{
  struct tm *p;
  static char buf[100];
  p = localtime(&t);
  strftime( buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", p);
  return buf;
}


/************************************************************************
 ************* CRC-32 subroutines ***************************************
 ************************************************************************/

uint CRCTab[256];

void InitCRC()
{
  for (int I=0;I<256;I++)
  {
    uint C=I;
    for (int J=0;J<8;J++)
      C=(C & 1) ? (C>>1)^0xEDB88320L : (C>>1);
    CRCTab[I]=C;
  }
}

uint UpdateCRC( void *Addr, uint Size, uint StartCRC)
{
  if (CRCTab[1]==0)
    InitCRC();
  uint8 *Data=(uint8 *)Addr;
#if defined(FREEARC_INTEL_BYTE_ORDER) && defined(PRESENT_UINT32)
  while (Size>=8)
  {
    StartCRC^=*(uint32 *)Data;
    StartCRC=CRCTab[(uint8)StartCRC]^(StartCRC>>8);
    StartCRC=CRCTab[(uint8)StartCRC]^(StartCRC>>8);
    StartCRC=CRCTab[(uint8)StartCRC]^(StartCRC>>8);
    StartCRC=CRCTab[(uint8)StartCRC]^(StartCRC>>8);
    StartCRC^=*(uint32 *)(Data+4);
    StartCRC=CRCTab[(uint8)StartCRC]^(StartCRC>>8);
    StartCRC=CRCTab[(uint8)StartCRC]^(StartCRC>>8);
    StartCRC=CRCTab[(uint8)StartCRC]^(StartCRC>>8);
    StartCRC=CRCTab[(uint8)StartCRC]^(StartCRC>>8);
    Data+=8;
    Size-=8;
  }
#endif
  for (int I=0;I<Size;I++)
    StartCRC=CRCTab[(uint8)(StartCRC^Data[I])]^(StartCRC>>8);
  return(StartCRC);
}

//  CRC  
uint CalcCRC( void *Addr, uint Size)
{
  return UpdateCRC (Addr, Size, INIT_CRC) ^ INIT_CRC;
}



//     from   to
char *replace (char *str, char* from, char to)
{
  for (char *p=str; *p; p++)
    if (in_set (*p, from))
      *p = to;
  return str;
}

//      
FILENAME basename (FILENAME fullname)
{
  char *basename = fullname;
  for (char* p=fullname; *p; p++)
    if (in_set (*p, ALL_PATH_DELIMITERS))
      basename = p+1;
  return basename;
}

//      name
void BuildPathTo (FILENAME name)
{
  char *path_ptr = NULL;
  for (char *p = str_end(name); --p >= name;)
    if (in_set (*p, DIRECTORY_DELIMITERS))
      {path_ptr=p; break;}
  if (path_ptr==NULL)  return;

  char oldc = *path_ptr;
  *path_ptr = '\0';

  if (! file_exists (name))
  {
    BuildPathTo (name);
    mkdir (name);
  }
  *path_ptr = oldc;
}

