/*
  FreeArc.CPP

  Second-level plugin module for FAR Manager 1.70 and MultiArc plugin

  Copyright (c) 1996-2000 Eugene Roshal
  Copyrigth (c) 2000-2002 FAR group
  Copyleft  (l) 2005      Bulat Ziganshin

  TO DO:  /    
         join_name - ,       
*/
/* Revision: 15.05.2005 $ */

#include <windows.h>
#include <string.h>
#include <dos.h>
#include <io.h>
#include <fcntl.h>
#include <stdio.h>
#include <time.h>
#include <setjmp.h>
#include <malloc.h>
#define _FAR_USE_FARFINDDATA
#include "plugin.hpp"
#include "fmt.hpp"

//       - longjmp       
jmp_buf jumper;
#define CHECK(a,b) {if(!(a)) longjmp(jumper,1);}
#include "ArcStructure.h"

/***********************************************************************
 ***   ************************************************
 ***********************************************************************
*/

static ARCHIVE *arcinfo;
static FILESIZE SFXSize;
static int current_block, current_file_in_block;
static DIRECTORY_BLOCK *dirblock;
static DWORD UnpVer;

// Number of 100 nanosecond units from 01.01.1601 to 01.01.1970
#define EPOCH_BIAS    116444736000000000ull

void WINAPI UnixTimeToFileTime( time_t time, FILETIME* ft )
{
  *(uint64*)ft = EPOCH_BIAS + time * 10000000ull;
}


BOOL WINAPI _export IsArchive(const char *Name,const unsigned char *Data,int DataSize)
{
  //    ,    -  
  for( int I=0; I <= (int)(DataSize-20); I++ )
  {
    if( *(uint32*)(Data+I) == aSIGNATURE)
    {
      UnpVer=(Data[I+4]*10+Data[I+5])*256 + Data[I+6]*10+Data[I+7];  //  ,    
      SFXSize = I;
      return TRUE;
    }
  }
  return FALSE;
}

BOOL WINAPI _export OpenArchive(const char *Name,int *Type)
{
  if (setjmp(jumper) != 0)
    return FALSE;  //           

  //   
  arcinfo = new ARCHIVE (FILENAME (Name));
  *Type   = 0;

  //  ,    GetArcItem     
  current_block=-1; dirblock=NULL;
  return TRUE;
}

int WINAPI _export GetArcItem(struct PluginPanelItem *Item,struct ArcItemInfo *Info)
{
  if (setjmp(jumper) != 0)
    return GETARC_BROKEN;  //           

  //     ,       
  if( current_block < 0 || ++current_file_in_block >= dirblock->total_files)
  {
    FreeAndNil (dirblock);
    for(;;)
    {
      if( ++current_block == arcinfo->control_blocks_descriptors.size )
      {
        return GETARC_EOF;
      }

      //     -      
      BLOCK& descriptor = arcinfo->control_blocks_descriptors [current_block];
      if (descriptor.type == DIR_BLOCK)
      {
         dirblock = new DIRECTORY_BLOCK (*arcinfo, descriptor);
         current_file_in_block = 0;
         if (dirblock->total_files>0)  break;
         FreeAndNil (dirblock);
      }
    }
    //printf("%d files\n", dirblock->total_files);
  }

  //   
  int i = current_file_in_block;
  Item->FindData.dwFileAttributes = dirblock->isdir[i]? FILE_ATTRIBUTE_DIRECTORY : 0;
  UnixTimeToFileTime( dirblock->time[i], &Item->FindData.ftLastWriteTime);
  Item->FindData.nFileSizeHigh = ((uint64) dirblock->size[i]) >> 32;
  Item->FindData.nFileSizeLow  = dirblock->size[i];
  dirblock->fullname (i, Item->FindData.cFileName);
  CharToOem( Item->FindData.cFileName, Item->FindData.cFileName );
  Item->CRC32  = dirblock->crc[i];
  Info->UnpVer = UnpVer;
  if (current_file_in_block==0)
  {
    uint64 packed = 0;
    iterate_array (j, dirblock->data_block)  packed += dirblock->data_block[j].compsize;
    Item->PackSizeHigh  = packed >> 32;
    Item->PackSize      = packed;
  }

  return GETARC_SUCCESS;
}

BOOL WINAPI _export CloseArchive(struct ArcInfo *Info)
{
  if (setjmp(jumper) != 0)
    return FALSE;  //           

  Info->SFXSize = SFXSize;
  Info->Comment = arcinfo->arcComment.size>0;
  Info->Lock    = arcinfo->arcLocked;
  FreeAndNil (dirblock);
  FreeAndNil (arcinfo);
  return TRUE;
}

DWORD WINAPI _export GetSFXPos(void)
{
  return SFXSize;
}


BOOL WINAPI _export GetFormatName(int Type,char *FormatName,char *DefaultExt)
{
  if (Type==0)
  {
    lstrcpy(FormatName,"FreeArc");
    lstrcpy(DefaultExt,"arc");
    return(TRUE);
  }
  return(FALSE);
}


BOOL WINAPI _export GetDefaultCommands(int Type,int Command,char *Dest)
{
  if (Type==0)
  {
    static char *Commands[]={
    /*Extract               */"arc x --noarcext -fn {%%S} -y -kb {-ap%%R} -- %%A @%%LNMA",
    /*Extract without paths */"arc e --noarcext -fn {%%S} -y -kb -- %%A @%%LNMA",
    /*Test                  */"arc t --noarcext -fn {%%S} -- %%A @%%LNMA",
    /*Delete                */"arc d --noarcext -fn {%%S} -- %%A @%%LNMA",
    /*Comment archive       */"arc c --noarcext -- %%A",
    /*Comment files         */"",
    /*Convert to SFX        */"arc s -- %%A",
    /*Lock archive          */"arc k --noarcext -- %%A",
    /*Protect archive       */"",
    /*Recover archive       */"",
    /*Add files             */"arc a {-ap%%R} --noarcext {%%S} -- %%A @%%LNMA",
    /*Move files            */"",
    /*Add files and folders */"arc a -r {-ap%%R} --noarcext {%%S} -- %%A @%%LNMA",
    /*Move files and folders*/"",
    /*"All files" mask      */"*"
    };
    if (Command < sizeof(Commands)/sizeof(Commands[0]))
    {
      lstrcpy(Dest,Commands[Command]);
      return(TRUE);
    }
  }
  return(FALSE);
}
